#include "fileio.h"
#include "card.h"
#include "cardfile.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

struct card_index
{
    long card_offset;
    char flag;

    char title[TITLE_WIDTH];
    int title_length;
};

enum cardfile_type { MGC, RRG, DKO };

/**************************** STATIC DECLARATIONS ****************************/
static int foreach_card_vp (struct card *cardfile, int (*f)(void *p, struct card *card), void *p);

static int write_mgc_header (FILE *f, struct card *cardfile);
static int write_indexes (FILE *f, struct card *cardfile);
static int write_one_index (FILE *p, struct card *c, long offset);
static int write_one_mgc_card (void *p, struct card *c);
static int bytes_one_mgc_card (void *p, struct card *c);

/* Returns success */
static int determine_type (FILE *f, enum cardfile_type *crd_type);
static int get_num_cards (FILE *f, enum cardfile_type crd_type, int *n);
static int read_card_indices (FILE *f, enum cardfile_type crd_type, struct card_index *cdx, int n);

/* Returns success */
static int read_card (FILE *f, enum cardfile_type crd_type, struct card_index *cdx, struct card *card);

/* Converts from little-endian to native-endian, returns success */
static int read_uint8 (FILE *f, unsigned char *uc);
static int read_uint16 (FILE *f, unsigned short *us);
static int read_uint32 (FILE *f, unsigned long *ul);

/* Converts from native-endian to little-endian, returns success */
static int write_uint8_fill (FILE *f, unsigned char uc, int n);
static int write_uint8 (FILE *f, unsigned char uc);
static int write_uint16 (FILE *f, unsigned short us);
static int write_uint32 (FILE *f, unsigned long ul);

/* Returns success */
static int skip_bytes (FILE *f, int n);
static int read_bytes (FILE *f, char *s, int n);
static int write_bytes (FILE *f, char *s, int n);
static int seek_to (FILE *f, long n);

static int get_has_object (FILE *f, enum cardfile_type crd_type, long card_offset, int *flag);
static int skip_presentation_object (FILE *f);

/* Returns meaningless offset if there is no text, returns success */
static int get_text_offset (FILE *f, enum cardfile_type crd_type, long card_offset, long *offset);

/* Returns meaningful offset even if there is no text, returns success */
static int get_text_len_offset (FILE *f, enum cardfile_type crd_type, long card_offset, long *offset);
/**************************** STATIC DECLARATIONS ****************************/


int write_cardfile (FILE *f, struct card *cardfile)
{
    int ret = write_mgc_header (f, cardfile) &&
    write_indexes (f, cardfile) &&
    foreach_card_vp (cardfile, write_one_mgc_card, f);

    if (ret) clear_all_dirty (cardfile);

    return ret;
}

struct card *read_cardfile (FILE *f)
{
    enum cardfile_type crd_type;
    int n, i;
    struct card_index *cdx;
    struct card *cardfile, *card;

    if (! determine_type (f, &crd_type) || ! get_num_cards (f, crd_type, &n) ||
        (cdx = malloc ((size_t) n * sizeof (struct card_index))) == NULL)
        return NULL;

    if (! read_card_indices (f, crd_type, cdx, n) ||
        (cardfile = alloc_blank_cardfile (n)) == NULL)
    {
        free (cdx);
        return NULL;
    }

    for (i = 0, card = cardfile; i < n; i++, card = card->next)
        if (! read_card (f, crd_type, &cdx[i], card)) 
        {
            delete_cardfile (cardfile);
            free (cdx);
            return NULL;
        }

    assert (n == count_cards (cardfile));

    free (cdx);
    return cardfile;
}

static int foreach_card_vp (struct card *cardfile, int (*f)(void *p, struct card *card), void *p)
{
    int ret = 1;

    cardfile = head (cardfile);

    while (cardfile)
    {
        ret = ret && f (p, cardfile);
        cardfile = cardfile->next;
    }

    return ret;
}

static int write_mgc_header (FILE *f, struct card *cardfile)
{
    return seek_to (f, 0) &&
    write_uint8 (f, '\x4d') &&
    write_uint8 (f, '\x47') &&
    write_uint8 (f, '\x43') &&
    write_uint16 (f, (unsigned short) count_cards (head (cardfile)));
}

static int write_indexes (FILE *f, struct card *cardfile)
{
    long offset;
    int ret = 1;

    cardfile = head (cardfile);
    offset = 5 + 52 * count_cards (cardfile);

    while (cardfile)
    {
        ret = ret && write_one_index (f, cardfile, offset);
        offset += bytes_one_mgc_card (f, cardfile);
        cardfile = cardfile->next;
    }

    return ret;
}

static int write_one_index (FILE *f, struct card *c, long offset)
{
    assert (TITLE_WIDTH == 40);

    return
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint8  (f, '\x00') &&
    write_uint32 (f, (unsigned long) offset) &&
    write_uint8  (f, '\x00') &&
    write_bytes (f, c->title->string, c->title->length) &&
    write_uint8_fill (f, '\0', TITLE_WIDTH - c->title->length) &&
    write_uint8  (f, '\x00');
}

/* Only writes text-only cards */
static int write_one_mgc_card (void *f, struct card *c)
{
    return
    write_uint16 (f, 0) &&
    write_uint16 (f, (unsigned short) c->body->text_length) &&
    write_bytes (f, c->body->text, c->body->text_length);
}

static int bytes_one_mgc_card (void *p, struct card *c)
{
    return 2 + 2 + c->body->text_length;
}

static int write_uint8_fill (FILE *f, unsigned char uc, int n)
{
    while (n--)
        if (! write_uint8 (f, uc))
            return 0;

    return 1;
}

static int read_uint8 (FILE *f, unsigned char *uc)
{
    if (fread (uc, 1, 1, f) != 1)
        return 0;

    return ! feof (f);
}

static int read_uint16 (FILE *f, unsigned short *us)
{
    unsigned char uc;

    if (fread (&uc, 1, 1, f) != 1)
        return 0;
    *us = uc;

    if (fread (&uc, 1, 1, f) != 1)
        return 0;
    *us += 256 * uc;

    return ! feof (f);
}

static int read_uint32 (FILE *f, unsigned long *ul)
{
    unsigned char uc;

    if (fread (&uc, 1, 1, f) != 1)
        return 0;
    *ul = uc;

    if (fread (&uc, 1, 1, f) != 1)
        return 0;
    *ul += 256 * uc;

    if (fread (&uc, 1, 1, f) != 1)
        return 0;
    *ul += 65536L * uc;

    if (fread (&uc, 1, 1, f) != 1)
        return 0;
    *ul += 16777216L * uc;

    return ! feof (f);
}

static int write_uint8 (FILE *f, unsigned char uc)
{
    return fwrite (&uc, 1, 1, f) == 1;
}

static int write_uint16 (FILE *f, unsigned short us)
{
    unsigned char uc;
    size_t written;

    uc = (unsigned char) (us % 256);
    written = fwrite (&uc, 1, 1, f);

    uc = (unsigned char) (us / 256);
    written += fwrite (&uc, 1, 1, f);

    return written == 2;
}

static int write_uint32 (FILE *f, unsigned long ul)
{
    unsigned char uc;
    size_t written;

    uc =  (unsigned char) (ul % 256);
    written = fwrite (&uc, 1, 1, f);

    uc = (unsigned char) ((ul / 256) % 256);
    written += fwrite (&uc, 1, 1, f);

    uc = (unsigned char) ((ul / 65536L) % 256);
    written += fwrite (&uc, 1, 1, f);

    uc =  (unsigned char) (ul / 16777216L);
    written += fwrite (&uc, 1, 1, f);

    return written == 4;
}

static int skip_bytes (FILE *f, int n)
{
    return ! fseek (f, n, SEEK_CUR);
}

static int read_bytes (FILE *f, char *s, int n)
{
    if (fread (s, 1, (size_t) n, f) != n)
        return 0;

    return ! feof (f);
}

static int write_bytes (FILE *f, char *s, int n)
{
    return fwrite (s, 1, (size_t) n, f) == n;
}

static int seek_to (FILE *f, long n)
{
    return ! fseek (f, n, SEEK_SET);
}

static int determine_type (FILE *f, enum cardfile_type *crd_type)
{
    char signature[4];

    seek_to (f, 0);
    read_bytes (f, signature, 3);
    signature[3] = '\0';

    if (strcmp (signature, "MGC") == 0)
    	*crd_type = MGC;
    else if (strcmp (signature, "RRG") == 0)
    	*crd_type = RRG;
    else if (strcmp (signature, "DKO") == 0)
    	*crd_type = DKO;
    else return 0;

    return 1;
}

static int get_num_cards (FILE *f, enum cardfile_type crd_type, int *n)
{
    int ret;
    unsigned short us = 0;

    switch (crd_type)
    {
    case RRG:
    case DKO:
    	ret = seek_to (f, 7) && read_uint16 (f, &us);
        *n = us;
        return ret;
    case MGC:
    	ret = seek_to (f, 3) && read_uint16 (f, &us);
        *n = us;
        return ret;
    default:
    	return 0;
    }
}

static int read_card_indices (FILE *f, enum cardfile_type crd_type, struct card_index *cdx, int n)
{
    int i, ret;

    /* DKO not supported yet */
    if (crd_type != MGC && crd_type != RRG)
    	return 0;

    if (! seek_to (f, crd_type == MGC ? 5 : 9))
    	return 0;

    ret = 1;

    for (i = 0; i < n; i++)
    {
    	char t[TITLE_WIDTH + 1];
        int j;

    	ret = ret && skip_bytes (f, 6);
    	ret = ret && read_uint32 (f, (unsigned long *) &cdx[i].card_offset);
    	ret = ret && read_uint8 (f, (unsigned char *) &cdx[i].flag);
    	ret = ret && read_bytes (f, cdx[i].title, TITLE_WIDTH);
    	ret = ret && skip_bytes (f, 1);

    	memcpy (t, cdx[i].title, TITLE_WIDTH);
    	t[TITLE_WIDTH] = '\0';

        /* Count from the end to preserve text after embedded NULs */
        for (j = TITLE_WIDTH - 1; j >= 0 && t[j] == '\0'; j--);
    	cdx[i].title_length = j + 1;
    }

    return ret;
}

static int get_has_object (FILE *f, enum cardfile_type crd_type, long card_offset, int *flag)
{
    unsigned short bitmap_length;

    /* First uint16 indicates bitmap length for MGC and boolean has_object for RRG */
    if ((crd_type == MGC || crd_type == RRG || crd_type != DKO) &&
        seek_to (f, card_offset) && read_uint16 (f, &bitmap_length))
    {
    	*flag = ! ! bitmap_length;
    	return 1;
    }

    return 0;
}

static int get_text_offset (FILE *f, enum cardfile_type crd_type, long card_offset, long *offset)
{
    if (crd_type == MGC)
    {
    	unsigned short bitmap_length;

    	if (! seek_to (f, card_offset) || ! read_uint16 (f, &bitmap_length))
    		return 0;

    	*offset = card_offset + (bitmap_length == 0 ? 4 : bitmap_length + 12);

    	return 1;
    }
    else if (crd_type == RRG)
    {
        unsigned short has_ole;

    	if (! seek_to (f, card_offset) || ! read_uint16 (f, &has_ole))
    		return 0;

        if (has_ole)
        {
            unsigned long ole_ver;
            unsigned long ole_format, l;

            /* Obj ID */
            if (! read_uint32 (f, &l))
                return 0;

            /* 1 for nt 3.51, 1.5 for win 3.1 */
            if (! read_uint32 (f, &ole_ver) ||
                (ole_ver != 0x0001 && ole_ver != 0x0501))
                abort ();

            if (! read_uint32 (f, &ole_format) || ! read_uint32 (f, &l) ||
                ! skip_bytes (f, (int) l))
                return 0;

            if (ole_format == 3)
            {
                /* Static object */
                if (! read_uint32 (f, &l) || ! read_uint32 (f, &l) ||
                    ! read_uint32 (f, &l) || ! skip_bytes (f, (int) l))
                    return 0;
            }
            else
            {
                /* Linked or embedded */
                if (! read_uint32 (f, &l) || ! skip_bytes (f, (int) l) ||
                    ! read_uint32 (f, &l) || ! skip_bytes (f, (int) l))
                    return 0;
                if (ole_format == 2)
                {
                    /* Embedded */
                    if (! read_uint32 (f, &l) || ! skip_bytes (f, (int) l))
                        return 0;
                    if (! skip_presentation_object (f))
                        return 0;
                }
                else
                {
                    /* Linked */
                    if (! read_uint32 (f, &l) || ! skip_bytes (f, (int) l) ||
                        ! read_uint32 (f, &l) || ! read_uint32 (f, &l))
                        return 0;
                    if (! skip_presentation_object (f))
                        return 0;
                }
            }

            /* Nine uint16 - char width & height, left, top, right, bottom, obj type */
            *offset = ftell (f) + 16;
        }
        else
            *offset = card_offset + 4;

    	return 1;
    }
    else
    	return 0;
}

static int skip_presentation_object (FILE *f)
{
    unsigned long l, ole_ver;
    unsigned char c;
    int i, gfx;

    /* ole version */
    if (! read_uint32 (f, &ole_ver) ||
        (ole_ver != 0x0001 && ole_ver != 0x0501))
        abort ();

    /* format */
    if (! read_uint32 (f, &l))
        return 0;

    /* No presentation object */
    if (l == 0)
        return 1;

    if (! read_uint32 (f, &l) || ! read_uint8 (f, &c))
        return 0;

    gfx = 0;

    /* Try to match METAFILEPICT\0 or BITMAP\0 or DIB\0 */
    switch (c)
    {
    case 'M':
        for (i = 0; i <= 11; i++)
        {
            if (! read_uint8 (f, &c))
                return 0;

            if (c != "ETAFILEPICT"[i])
            {
                while (c)
                    if (! read_uint8 (f, &c))
                        return 0;
                break;
            }
        }
        gfx = 1;
        break;

    case 'B':
        for (i = 0; i <= 5; i++)
        {
            if (! read_uint8 (f, &c))
                return 0;

            if (c != "ITMAP"[i])
            {
                while (c)
                    if (! read_uint8 (f, &c))
                        return 0;
                break;
            }
        }
        gfx = 1;
        break;

    case 'D':
        for (i = 0; i <= 2; i++)
        {
            if (! read_uint8 (f, &c))
                return 0;

            if (c != "IB"[i])
            {
                while (c)
                    if (! read_uint8 (f, &c))
                        return 0;
                break;
            }
        }
        gfx = 1;
        break;
    }

    if (gfx)
    {
        if (! read_uint32 (f, &l) || ! read_uint32 (f, &l) ||
            ! read_uint32 (f, &l) || ! skip_bytes (f, (int) l))
            return 0;
    }
    else
    {
        if (! read_uint32 (f, &l))
            return 0;
        if (! l)
            if (! read_uint32 (f, &l) || ! skip_bytes (f, (int) l))
                return 0;
        if (! read_uint32 (f, &l) || ! skip_bytes (f, (int) l))
            return 0;
    }

    return 1;
}

static int get_text_len_offset (FILE *f, enum cardfile_type crd_type, long card_offset, long *offset)
{
    /* in all three cardfile types, the text length is the uint16
       preceeding the text itself */

    if (! get_text_offset (f, crd_type, card_offset, offset))
    	return 0;

    *offset -= 2;
    return 1;
}

static int read_card (FILE *f, enum cardfile_type crd_type, struct card_index *cdx, struct card *card)
{
    int has_object, ret;
    long text_offset, text_len_offset;
    unsigned short us = 0;

    if (! cdx || ! card ||
        ! get_has_object      (f, crd_type, cdx->card_offset, &has_object) ||
        ! get_text_offset     (f, crd_type, cdx->card_offset, &text_offset) ||
        ! get_text_len_offset (f, crd_type, cdx->card_offset, &text_len_offset))
        return 0;

    editstring_set_cstr_with_length (card->title, cdx->title, cdx->title_length);

    /* We strip out the objects -- alert user before save */
    card->object_skipped = has_object;

    ret = seek_to (f, text_len_offset) &&
          read_uint16 (f, &us);
    card->body->text_length = us;

    return ret && seek_to (f, text_offset) &&
                  read_bytes (f, card->body->text, card->body->text_length);
}

